#include "serial.hpp"
#include "../uartprint.hpp"
#include "rom/md5_hash.h"
#include <cstdio>
#include <dirent.h>
#include "../utils/ensurerange.hpp"

// Maximum length: 988

SerialBluetooth* global_serial_bluetooth = nullptr;

SerialBluetooth::SerialBluetooth(VFS* vfs, MotorControl* mc){
    if(global_serial_bluetooth != nullptr){
        throw "Only one instance is allowed.";
    }
    this->vfs = vfs;
    this->mc = mc;
    this->bs = new BluetoothSerial();
    String bt_name("Smart Tail");
    auto* tailname = vfs->nvs->get_string("tailname");
    if(tailname!=nullptr){
        bt_name.concat(": ");
        bt_name.concat(tailname->array);
        delete tailname;
    }
    if(!this->bs->begin(bt_name)){
        uart_print("Bluetooth failed to initialize!\n");
        throw "BT_init error";
    }
    global_serial_bluetooth = this;
    this->bs->register_callback(bluetooth_callback);
}

void SerialBluetooth::send(char str){
    this->bs->write(str);
}

void SerialBluetooth::send(char* str){
    this->bs->write(reinterpret_cast<const uint8_t*>(str), strlen(str));
}

void SerialBluetooth::send(const char* str){
    this->bs->write(reinterpret_cast<const uint8_t*>(str), strlen(str));
}

void SerialBluetooth::send(std::string str){ this->send(str.c_str()); }
void SerialBluetooth::send(String str){ this->send(str.c_str()); }
void SerialBluetooth::send(uint8_t* str){ this->send(reinterpret_cast<char*>(str)); }
void SerialBluetooth::send(int8_t* str){ this->send(reinterpret_cast<char*>(str)); }

void SerialBluetooth::_send_pending_stuff_value_from_nvs(std::string name){
    auto* attr = this->vfs->nvs->get_string(name);
    if(attr==nullptr){
        this->send(name);
        this->send(":\n");
        this->send("missing:");
        this->send(name);
        this->send('\n');
    } else {
        this->send(name);
        this->send(':');
        this->send(attr->array);
        this->send('\n');
        delete attr;
    };
}

void SerialBluetooth::_send_pending_stuff(){
    this->_send_pending_stuff_value_from_nvs("tailname");
    this->_send_pending_stuff_value_from_nvs("min_l_pos");
    this->_send_pending_stuff_value_from_nvs("min_r_pos");
    this->_send_pending_stuff_value_from_nvs("max_l_pos");
    this->_send_pending_stuff_value_from_nvs("max_r_pos");
}

void SerialBluetooth::put_on_action_queue(Action* action){
    this->mc->motor_logic->mutex->acquire();
    this->mc->motor_logic->pending_actions->push(action);
    this->mc->motor_logic->mutex->release();
}

void SerialBluetooth::_on_incoming_message(String* message){
    if(message->startsWith("echo:")){
        this->send("reply");
        this->send(message->c_str());
    } else if (message->startsWith("set_param:")) {
        String param_val = message->substring(message->indexOf(':')+1);
        int sep = param_val.indexOf(':');
        String param = param_val.substring(0, sep);
        String val = param_val.substring(sep+1, param_val.length()-1);
        this->vfs->nvs->set_string(param.c_str(), val.c_str());
        this->put_on_action_queue(new ActionConfigReload());
        if(param.equals("default_rhyrhm")){
            int rhythm = atoi(val.c_str());
            if(rhythm<=0) rhythm = 90;
            this->put_on_action_queue(new ActionChangeRhythm(rhythm));
        }
    } else if (message->startsWith("set_rhythm:")) {
        String newrth = message->substring(message->indexOf(':')+1);
        int rhythm = atoi(newrth.c_str());
        if(rhythm<=0) rhythm = 90;
        this->put_on_action_queue(new ActionChangeRhythm(rhythm));
    } else if (message->startsWith("reengage")) {
        this->put_on_action_queue(new ActionReengage());
    } else if (message->startsWith("disengage")) {
        this->put_on_action_queue(new ActionDisengage());
    } else if (message->startsWith("status_1")) {
        this->_send_pending_stuff();
    } else if (message->startsWith("status_2")) {
        this->send("current_l:");
        this->send(String(this->mc->left_stepper->stepper_motor->get_current()).c_str());
        this->send('\n');
        this->send("current_r:");
        this->send(String(this->mc->right_stepper->stepper_motor->get_current()).c_str());
        this->send('\n');
        this->send("target_l:");
        this->send(String(this->mc->left_stepper->stepper_motor->get_target()).c_str());
        this->send('\n');
        this->send("target_r:");
        this->send(String(this->mc->right_stepper->stepper_motor->get_target()).c_str());
        this->send('\n');
    } else if (message->startsWith("status_3")) {
        this->send("engaged_l:");
        this->send(this->mc->left_stepper->get_engaged() ? 'y' : 'n');
        this->send('\n');
        this->send("engaged_r:");
        this->send(this->mc->right_stepper->get_engaged() ? 'y' : 'n');
        this->send('\n');
        this->send("engaging_l:");
        this->send(this->mc->left_stepper->get_engaging() ? 'y' : 'n');
        this->send('\n');
        this->send("engaging_r:");
        this->send(this->mc->right_stepper->get_engaging() ? 'y' : 'n');
        this->send('\n');
    } else if (message->startsWith("status_4")) {
        size_t total_bytes;
        size_t used_bytes;
        esp_spiffs_info(NULL, &total_bytes, &used_bytes);
        this->send("total_bytes:");
        this->send(String(total_bytes).c_str());
        this->send('\n');
        this->send("used_bytes:");
        this->send(String(used_bytes).c_str());
        this->send('\n');
        int fc = 0;
        DIR* dp;
        struct dirent* dep;
        dp = opendir("/spiffs");
        while ((dep = readdir(dp)) != NULL) {
            if (
                String("multipart~").equals(dep->d_name)
                ||
                String("idx~").equals(dep->d_name)
            ){}else{
                fc++;
            }
        }
        this->send("files:");
        this->send(String(fc).c_str());
        this->send('\n');
    } else if (message->startsWith("filelist")) {
        std::vector<std::string> files;
        int fc = 0;
        DIR* dp;
        struct dirent* dep;
        dp = opendir("/spiffs");
        while ((dep = readdir(dp)) != NULL) {
            if (
                String("multipart~").equals(dep->d_name)
                ||
                String("idx~").equals(dep->d_name)
            ){}else{
                fc++;
                files.push_back(String(dep->d_name).c_str());
            }
        }
        std::sort(files.begin(), files.end());
        int filepos = -1;
        if(message->indexOf(':')>=0){
            filepos = atoi(message->substring(message->indexOf(':')+1).c_str());
        }
        if(filepos<0 or filepos>=files.size()){
            this->send("files:");
            this->send(String(fc).c_str());
            this->send('\n');
        } else {
            this->send("file#");
            this->send(String(filepos).c_str());
            this->send(":");
            this->send(files.at(filepos));
            this->send('\n');
        }
    } else if (message->startsWith("multipart_session_start:")) {
        String filename = message->substring(message->indexOf(':')+1);
        if(
            filename.equals("idx~\n")
            ||
            filename.equals("multipart~\n")
        ){
            this->send("error:multipart_session_start:");
            this->send("forbidden filename");
            this->send('\n');
        } else {
            FILE* multipart = fopen("/spiffs/multipart~", "w");
            fwrite(filename.c_str(), sizeof(char), filename.length(), multipart);
            fclose(multipart);
            this->send("multipart_session_start:");
            this->send(filename);
            this->send('\n');
        }
    } else if (message->startsWith("multipart_session_append:")) {
        String content = message->substring(message->indexOf(':')+1, message->length()-1);
        FILE* multipart = fopen("/spiffs/multipart~", "a");
        fwrite(content.c_str(), sizeof(char), content.length(), multipart);
        fclose(multipart);
        this->send("multipart_session_start:");
        this->send(String(content.length()).c_str());
        this->send('\n');
    } else if (message->startsWith("multipart_session_consolidate")) {
        FILE* multipart = fopen("/spiffs/multipart~", "r");
        fseek(multipart, 0L, SEEK_END);
        int mpsize = ftell(multipart);
        char* mpcontents_ = (char*) calloc(mpsize+1, sizeof(char));
        fseek(multipart, 0L, SEEK_SET);
        fread(mpcontents_, sizeof(char), mpsize, multipart);
        String mpcontents(mpcontents_);
        free(mpcontents_);
        fclose(multipart);
        int firstnl = mpcontents.indexOf("\n");
        String filename = mpcontents.substring(0, firstnl);
        String filecontents = mpcontents.substring(firstnl+1, mpcontents.length());
        String path("/spiffs/");
        path.concat(filename);
        FILE* fp = fopen(path.c_str(), "w");
        fwrite(filecontents.c_str(), sizeof(char), filecontents.length(), fp);
        fclose(fp);
        unlink("/spiffs/multipart~");
        this->send("created:");
        this->send(filename);
        this->send('\n');
    } else if (message->startsWith("multipart_session_abort")) {
        unlink("/spiffs/multipart~");
        this->send("deleted:multipart~");
        this->send('\n');
    } else if (message->startsWith("delete:")) {
        String filename = message->substring(message->indexOf(':')+1, message->length()-1);
        String path("/spiffs/");
        path.concat(filename);
        unlink(path.c_str());
        this->send("deleted:");
        this->send(filename);
        this->send('\n');
    } else if (message->startsWith("play:")) {
        String filename = message->substring(message->indexOf(':')+1, message->length()-1);
        String path("/spiffs/");
        path.concat(filename);
        FILE* fp = fopen(path.c_str(), "r");
        if (fp!=NULL){
            std::string contents;
            {
                fseek(fp, 0L, SEEK_END);
                int mpsize = ftell(fp);
                char* contents_ = (char*) calloc(mpsize+1, sizeof(char));
                fseek(fp, 0L, SEEK_SET);
                fread(contents_, sizeof(char), mpsize, fp);
                fclose(fp);
                String cont(contents_);
                free(contents_);
                contents = cont.c_str();
            }
            TimelineCoreographer* coreographer = nullptr;
            try{
                coreographer = new TimelineCoreographer(contents);
            }catch(std::string exc){
                this->send("error:");
                this->send("corrupted");
                this->send('\n');
                this->send("error");
                this->send(exc);
                this->send('\n');
                Serial.println(exc.c_str());
            };
            if(coreographer!=nullptr){
                this->put_on_action_queue(new ActionSetCoreographer(coreographer));
                this->send("playing:");
                this->send(filename);
                this->send('\n');
            }
        }else{
            this->send("error:");
            this->send("does not exist");
            this->send('\n');
        }
    } else if (message->startsWith("goto_abs:")) {
        String lr = message->substring(message->indexOf(':')+1, message->length()-1);
        String ls = lr.substring(0, lr.indexOf(':'));
        String rs = lr.substring(lr.indexOf(':')+1);
        int l = atoi(ls.c_str());
        int r = atoi(rs.c_str());
        this->put_on_action_queue(new ActionSetCoreographer(new CoreographerHoldingAbsolutePosition(l, r)));
        this->send("gone_abs\n");
    } else if (message->startsWith("goto_rel:")) {
        String lr = message->substring(message->indexOf(':')+1, message->length()-1);
        String ls = lr.substring(0, lr.indexOf(':'));
        String rs = lr.substring(lr.indexOf(':')+1);
        float l = atof(ls.c_str());
        float r = atof(rs.c_str());
        if(std::isnan(l) || std::isinf(l)) l = 0;
        if(std::isnan(r) || std::isinf(r)) r = 0;
        l = ensure_range(0.0f, l, 1.0f);
        r = ensure_range(0.0f, r, 1.0f);
        this->put_on_action_queue(new ActionSetCoreographer(new CoreographerHoldingPosition(l, r)));
        this->send("gone_rel\n");
    } else {
        this->send("unknown_cmd:");
        this->send(message->c_str());
    }
}

void bluetooth_callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param){
    if(global_serial_bluetooth==nullptr) return;
    auto* bt = global_serial_bluetooth;
    if(event == ESP_SPP_SRV_OPEN_EVT){
        String client_mac("");
        uart_print("Incoming connection from: ");
        for(int i=0; i<6; i++) {
            char hex[3] = {'0', '0', '\0'};
            sprintf(hex, "%02X", param->srv_open.rem_bda[i]);
            client_mac.concat(hex);
            if(i<5) client_mac.concat(':');
        }
        uart_print(client_mac.c_str());
        uart_print("\n");
        auto* kbtm_a = bt->vfs->nvs->get_string("bluetooth_mac");
        String known_bluetooth_mac((kbtm_a==nullptr) ? "" : kbtm_a->array);
        if(kbtm_a!=nullptr) delete kbtm_a;
        if(known_bluetooth_mac.equals("")){
            bt->vfs->nvs->set_string("bluetooth_mac", client_mac.c_str());
            known_bluetooth_mac = client_mac;
        }
        if(known_bluetooth_mac.equals(client_mac.c_str())){
            bt->send("welcome:");
            bt->send(client_mac);
            bt->send("\n");
            uart_print("Client Connected\n");
            bt->_send_pending_stuff();
            fclose(fopen("/spiffs/idx~", "a+"));
        } else {
            esp_spp_disconnect(param->open.handle);
            uart_print("Client Rejected\n");
        }
    } else if (event == ESP_SPP_CLOSE_EVT){
        uart_print("Client Disonnected\n");
    } else if (event == ESP_SPP_DATA_IND_EVT) {
        auto* arr = new Array<uint8_t>(param->data_ind.len+1);
        memcpy(arr->array, param->data_ind.data, param->data_ind.len);
        String* str = new String((char*) arr->array);
        delete arr;
        {
            String md5hex("");
            {
                MD5Context* md5ctx = (MD5Context*) calloc(1, sizeof(MD5Context));
                MD5Init(md5ctx);
                MD5Update(md5ctx, param->data_ind.data, param->data_ind.len);
                {
                    unsigned char md5digest[16] = {0};
                    MD5Final(md5digest, md5ctx);
                    for(int i=0; i<16; i++){
                        char hex[3] = {'0', '0', '\0'};
                        sprintf(hex, "%02X", md5digest[i]);
                        md5hex.concat(hex);
                    }
                }
                free(md5ctx);
            }
            bt->send("received:");
            bt->send(md5hex);
            bt->send('\n');
        }
        bt->_on_incoming_message(str);
        delete str;
    }
    
}

void SerialBluetooth::mainloop(){
    while (this->bs->available()) {
        this->bs->read();
    }
}
