#include "coreographer.hpp"
#include "limits.h"
#include <cmath>
#include <cstdlib>
#include <algorithm>
#include "../utils/explode.hpp"

bool tecmp(const std::unique_ptr<TimelineElement>& a, const std::unique_ptr<TimelineElement>& b){ return (a->t) < (b->t); }

Coreographer::Coreographer(){ this->update_last_tick(); }
int64_t Coreographer::get_current_tick(){ return esp_timer_get_time(); }
int64_t Coreographer::get_last_tick(){ return this->last_tick; }
int64_t Coreographer::update_last_tick(){ return this->last_tick = this->get_current_tick(); }
int64_t Coreographer::since_last_tick(bool updating){
    int64_t lt = this->last_tick;
    int64_t ct = this->get_current_tick();
    if(updating) this->last_tick = ct;
    if(ct<lt)
        return (ct-LONG_MIN)+(LONG_MAX-lt);
    else
        return ct - lt;
}

void Coreographer::set_boundaries(int minl, int maxl, int minr, int maxr){
    this->minl=minl;
    this->maxl=maxl;
    this->minr=minr;
    this->maxr=maxr;
}

CoreographerHoldingPosition::CoreographerHoldingPosition(float l, float r): Coreographer(){
    this->target_l = l;
    this->target_r = r;
}

void CoreographerHoldingPosition::setup_stepper(SmartStepper* l_stepper, SmartStepper* r_stepper, int rhythm){
    this->since_last_tick();
    l_stepper->set_target(round(this->minl + (this->target_l*(this->maxl-this->minl))));
    r_stepper->set_target(round(this->minr + (this->target_r*(this->maxr-this->minr))));
}

CoreographerHoldingAbsolutePosition::CoreographerHoldingAbsolutePosition(int l, int r): Coreographer(){
    this->target_l = l;
    this->target_r = r;
}

void CoreographerHoldingAbsolutePosition::setup_stepper(SmartStepper* l_stepper, SmartStepper* r_stepper, int rhythm){
    this->since_last_tick();
    l_stepper->set_target(this->target_l);
    r_stepper->set_target(this->target_r);
}

TimelineCoreographer::TimelineCoreographer(std::string whole_marshalled){
    std::vector<std::string> whole_structured = explode(whole_marshalled, '@');
    if(whole_structured.size()!=2) throw "@tech: cannot be divided exactly in head and tail with '@'";
    std::string head_marshalled = whole_structured.at(0);
    std::vector<std::string> head_structured = explode(head_marshalled, '$');
    if(head_structured.size()!=3) throw "@tech: head should be divided in 3 elements using '$'";
    int event_count = atoi(head_structured.at(0).c_str());
    if(event_count<1) throw "@tech: head[0] should be at least 1";
    int bpm_divider = atoi(head_structured.at(1).c_str());
    if(bpm_divider<1) throw "@tech: head[1] should be at least 1";
    float cycle_window = atof(head_structured.at(2).c_str());
    if(std::isnan(cycle_window) || std::isinf(cycle_window)) throw "@tech: head[2] should not be NaN or Inf";
    if(cycle_window>0) throw "@tech: head[2] should be greater than zero";
    std::string body_marshalled = whole_structured.at(1);
    std::vector<std::string> body_structured = explode(body_marshalled, '#');
    if(body_structured.size()!=event_count) throw "@tech: head[0] does not reflect body size divided with '#'";
    this->elements = std::vector<std::unique_ptr<TimelineElement>>();
    for(std::string element_marshalled: body_structured){
        std::vector<std::string> element_structured = explode(element_marshalled, '$');
        if(element_structured.size()!=4) throw "@tech: body[?] should have exactly 4 elements";
        float window_at = atof(element_structured.at(0).c_str());
        if(std::isnan(window_at) || std::isinf(window_at)) throw "@tech: body[?][0] should not be NaN or Inf";
        if(!(0<=window_at && window_at<=cycle_window)) throw "@tech: body[?][0] should be between 0 and head[2]";
        float pos_stp_l = atof(element_structured.at(1).c_str());
        if(std::isnan(pos_stp_l) || std::isinf(pos_stp_l)) throw "@tech: body[?][1] should not be NaN or Inf";
        if(!(0<=pos_stp_l && pos_stp_l<=1)) throw "@tech: body[?][1] should be between 0 and 1";
        float pos_stp_r = atof(element_structured.at(2).c_str());
        if(std::isnan(pos_stp_r) || std::isinf(pos_stp_r)) throw "@tech: body[?][2] should not be NaN or Inf";
        if(!(0<=pos_stp_r && pos_stp_r<=1)) throw "@tech: body[?][2] should be between 0 and 1";
        std::string interpolation_marshalled = element_structured.at(3);
        if(interpolation_marshalled.size()<=0) throw "@tech: body[?][3] should not be empty";
        char interpolation_key = interpolation_marshalled.at(0);
        std::string interpolation_arguments_marshalled = interpolation_marshalled.substr(1);
        std::vector<std::string> interpolation_arguments_structured = (
            (interpolation_arguments_marshalled.size() > 0)
                ? explode(interpolation_arguments_marshalled, '%')
                : std::vector<std::string>()
        );
        Interpolation* interpolation = get_interpolation(interpolation_key, interpolation_arguments_structured);
        this->elements.push_back(std::unique_ptr<TimelineElement>(new TimelineElement(window_at/cycle_window, pos_stp_l, pos_stp_r, interpolation)));
    }
    this->bpm_divider = bpm_divider;
    std::stable_sort(this->elements.begin(), this->elements.end(), tecmp);
}

double put_in_place(double a){
    while(a>=1.0)
        a -= 1.0;
    return a;
}

void TimelineCoreographer::setup_stepper(SmartStepper* l_stepper, SmartStepper* r_stepper, int rhythm_){
    int64_t rhythm = static_cast<int64_t>((static_cast<double>(rhythm_)/(this->bpm_divider*60))*1000*1000);
    int64_t slt = this->since_last_tick();
    this->normalized_position += static_cast<double>(slt)/rhythm;
    this->normalized_position = put_in_place(this->normalized_position);
    double np = this->normalized_position;
    TimelineElement* first = this->elements.at(0).get();
    TimelineElement* last = this->elements.at(this->elements.size()-1).get();
    float l = 0;
    float r = 0;
    if ((np > last->t) || (np < first->t)){
        double r = put_in_place(np - last->t + 1.0) / put_in_place(first->t - last->t + 1.0);
        l = last->i->interpolate(last->l, first->l, r);
        r = last->i->interpolate(last->r, first->r, r);
    } else {
        TimelineElement* begin = first;
        TimelineElement* end = last;
        for(std::unique_ptr<TimelineElement>& upte: this->elements){
            TimelineElement* te = upte.get();
            if(begin->t < te->t && te->t <= np)
                begin = te;
            if(np <= te->t && te->t < end->t)
                end = te;
        }
        double r = put_in_place(np - begin->t + 1.0) / put_in_place(end->t - begin->t + 1.0);
        l = begin->i->interpolate(begin->l, end->l, r);
        r = begin->i->interpolate(begin->r, end->r, r);
    }
    l_stepper->set_target(round(l));
    r_stepper->set_target(round(r));
}

TimelineElement::TimelineElement(float window_at, float pos_stp_l, float pos_stp_r, Interpolation* interpolation){
    this->t = window_at;
    this->l = pos_stp_l;
    this->r = pos_stp_r;
    this->i = std::unique_ptr<Interpolation>(interpolation);
}
