#include "interpolation.hpp"
#include "../utils/ensurerange.hpp"
#include <cstdlib>
#include <cmath>

Interpolation* get_interpolation(char method, std::vector<std::string> arguments){
    switch(method){
        case 'a': return new InterpolationLinear(arguments); //linear in, linear out
        case 'b': return new InterpolationWeighted( 2,  2); //ease   in, ease   out
        case 'c': return new InterpolationWeighted(.5, .5); //harden in, harden out
        case 'd': return new InterpolationWeighted( 2, .5); //ease   in, harden out
        case 'e': return new InterpolationWeighted(.5,  2); //harden in, ease   out
        case 'f': return new InterpolationWeighted( 2,  1); //ease   in, linear out
        case 'g': return new InterpolationWeighted(.5,  1); //harden in, linear out
        case 'h': return new InterpolationWeighted( 1,  2); //linear in, ease   out
        case 'i': return new InterpolationWeighted( 1, .5); //linear in, harden out
        case 'j': return new InterpolationStep(0.0); //step release
        case 'k': return new InterpolationStep(0.5); //step middle
        case 'l': return new InterpolationStep(1.0); //step hold
        case 'y': return new InterpolationStep(arguments);
        case 'z': return new InterpolationWeighted(arguments);
        default: return new InterpolationLinear(arguments);
    }
}

InterpolationLinear::InterpolationLinear(std::vector<std::string> arguments): Interpolation(){}
InterpolationLinear::InterpolationLinear(): Interpolation(){}
float InterpolationLinear::interpolate(float before, float after, float middle){
    middle = ensure_range(0.0f, middle, 1.0f);
    float middle_b = 1.0f-middle;
    float middle_a = middle;
    return (before*middle_b) + (after*middle_a);
}

InterpolationWeighted::InterpolationWeighted(float i, float o): Interpolation(){
    this->weight_in = i;
    this->weight_out = o;
    if(this->weight_in<=0) throw "@tech: body[?][3].args[0] should be greater than zero";
    if(this->weight_out<=0) throw "@tech: body[?][3].args[1] should be greater than zero";
    if(std::isnan(this->weight_in)  || std::isinf(this->weight_in))  throw "@tech: body[?][3].args[0] should not be NaN or Inf";
    if(std::isnan(this->weight_out) || std::isinf(this->weight_out)) throw "@tech: body[?][3].args[1] should not be NaN or Inf";
}
InterpolationWeighted::InterpolationWeighted(std::vector<std::string> arguments): Interpolation(){
    if(arguments.size()!=2) throw "@tech: body[?][3].args should have exactly 2 elements"; 
    this->weight_in = atof(arguments.at(0).c_str());
    this->weight_out = atof(arguments.at(1).c_str());
    if(this->weight_in<=0) throw "@tech: body[?][3].args[0] should be greater than zero";
    if(this->weight_out<=0) throw "@tech: body[?][3].args[1] should be greater than zero";
    if(std::isnan(this->weight_in)  || std::isinf(this->weight_in))  throw "@tech: body[?][3].args[0] should not be NaN or Inf";
    if(std::isnan(this->weight_out) || std::isinf(this->weight_out)) throw "@tech: body[?][3].args[1] should not be NaN or Inf";
}
float InterpolationWeighted::interpolate(float before, float after, float middle){
    middle = ensure_range(0.0f, middle, 1.0f);
    float middle_b = pow(1.0f-middle, this->weight_in);
    float middle_a = pow(middle, this->weight_out);
    return ((before*middle_b) + (after*middle_a)) / (middle_b + middle_a);
}

InterpolationStep::InterpolationStep(float cutoff): Interpolation(){
    this->cutoff = ensure_range(0.0f, cutoff, 1.0f);
    if(std::isnan(this->cutoff) || std::isinf(this->cutoff)) throw "@tech: body[?][3].args[0] should not be NaN or Inf";
}
InterpolationStep::InterpolationStep(std::vector<std::string> arguments): Interpolation(){
    if(arguments.size()!=1) throw "@tech: body[?][3].args should have exactly 1 element"; 
    this->cutoff = ensure_range(0.0f, cutoff, 1.0f);
    if(std::isnan(this->cutoff) || std::isinf(this->cutoff)) throw "@tech: body[?][3].args[0] should not be NaN or Inf";
}
float InterpolationStep::interpolate(float before, float after, float middle){
    middle = ensure_range(0.0f, middle, 1.0f);
    return (middle<=this->cutoff) ? before : after;
}

Interpolation* InterpolationLinear::clone(){ return new InterpolationLinear(); }
Interpolation* InterpolationWeighted::clone(){ return new InterpolationWeighted(this->weight_in, this->weight_out); }
Interpolation* InterpolationStep::clone(){ return new InterpolationStep(this->cutoff); }
